/*
 * Musepack audio compression
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <math.h>
#include <string.h>

#include <mpc/mpc_types.h>
#include <mpc/datatypes.h>
#include <mpc/mpcmath.h>

#include "../include/libmpcenc.h"
#include "psy_model.h"
#include "bitstream.h"
#include "encoder.h"
#include "syntax_sv8.h"

#pragma warning(disable: 4244)	// warning C4244: '+=' : conversion from 'double' to 'float', possible loss of data

// constructor & destructor
int mpc_encoder_open(mpc_encoder_t *instance)
{
	mpc_enc_t	*encoder;

	if (!instance) return -1;

	// new instance of the encoder
	encoder = (mpc_enc_t*) malloc(sizeof(mpc_enc_t));
	if (!encoder) {
		return -1;
	}

	// everything is blank...
	memset(encoder, 0, sizeof(mpc_enc_t));

	*instance = (mpc_encoder_t)encoder;
	return 0;
}

int mpc_encoder_close(mpc_encoder_t instance)
{
	mpc_enc_t	*encoder = (mpc_enc_t*)instance;

	if (!instance) return 0;

	// release buffers... etc.
	if (encoder->buffer) {
		free(encoder->buffer);
		encoder->buffer = NULL;
	}

	free(encoder);
	return 0;
}

// simple way to configure the encoder
int mpc_encoder_config(mpc_encoder_t instance, int samplerate, int channels, int quality)
{
	mpc_enc_t				*encoder = (mpc_enc_t*)instance;
	mpc_encoder_config_t	*cfg	 = NULL;

	if (!encoder) return -1;

	cfg = &encoder->config;
	
	// verify parameters
	if (samplerate != 44100 &&
		samplerate != 48000 &&
		samplerate != 37800 &&
		samplerate != 32000) return -2;

	if (channels <= 0 ||
		channels > 2) return -3;

	if (quality < 0 ||
		quality > 10) return -4;

	// prepare default config with given quality
	cfg->samplerate				= samplerate;
	cfg->channels				= channels;
	cfg->quality				= quality;
	cfg->stream_version			= 8;
	
	cfg->enable_seek_table		= 0;
	cfg->seek_power				= 2;
	cfg->frames_per_block_power	= 6;

	// preliminary initialization
	mpc_psy_model_set_quality(&encoder->psy_model, quality);

	// copy the values from PSY model
	cfg->ns_order		 = encoder->psy_model.NS_Order;
	cfg->var_ltq		 = encoder->psy_model.varLtq;
	cfg->pns			 = encoder->psy_model.PNS;		
	cfg->min_val_choice	 = encoder->psy_model.MinValChoice;
	cfg->trans_detect	 = encoder->psy_model.TransDetect;
	cfg->short_thr		 = encoder->psy_model.ShortThr;
	cfg->nmt			 = encoder->psy_model.NMT;
	cfg->tmn			 = encoder->psy_model.TMN;
	cfg->cvd_used		 = encoder->psy_model.CVD_used;
	cfg->ms_channel_mode = encoder->psy_model.MS_Channelmode;
	cfg->min_smr		 = encoder->psy_model.minSMR;
	cfg->tmp_mask_used	 = encoder->psy_model.tmpMask_used;

	return 0;
}


int mpc_encoder_get_config(mpc_encoder_t instance, mpc_encoder_config_t **config)
{
	mpc_enc_t	*encoder = (mpc_enc_t*)instance;
	
	if (!encoder || !config) return -1;

	// pointer to our config structure
	*config = &encoder->config;
	return 0;
}


int mpc_encoder_init(mpc_encoder_t instance)
{
	mpc_enc_t				*encoder = (mpc_enc_t*)instance;
	mpc_encoder_config_t	*cfg	 = NULL;
	PsyModel				*m		 = NULL;
	
	if (!encoder) return -1;

	cfg = &encoder->config;
	m   = &encoder->psy_model;

	// now we need to calculate the buffer size
	cfg->buffer_size	= (1 << cfg->frames_per_block_power) * sizeof(mpc_uint8_t) * MAX_FRAME_SIZE;

	// alloc our own buffer
	if (encoder->buffer) {
		free(encoder->buffer);
		encoder->buffer = NULL;
	}
	encoder->buffer = (mpc_uint8_t*)malloc(cfg->buffer_size);
	if (!encoder->buffer) return -2;

	// let the encoder know
	encoder->frames_per_block_pwr	= cfg->frames_per_block_power;

	// reflect the changes into the PSY model
	m->NS_Order		  = cfg->ns_order;
	m->varLtq		  = cfg->var_ltq;
	m->PNS			  = cfg->pns;		
	m->MinValChoice   = cfg->min_val_choice;
	m->TransDetect	  = cfg->trans_detect;
	m->ShortThr		  = cfg->short_thr;
	m->NMT			  = cfg->nmt;
	m->TMN			  = cfg->tmn;
	m->CVD_used		  = cfg->cvd_used;
	m->MS_Channelmode = cfg->ms_channel_mode ;
	m->minSMR		  = cfg->min_smr;
	m->tmpMask_used   = cfg->tmp_mask_used;

	// also init the psy model
	mpc_psy_model_init(&encoder->psy_model);

	return 0;
}


// encoding PCM samples
int mpc_encoder_encode_pcm(mpc_encoder_t instance, float *samples, int count, mpc_uint8_t *buf, int *bytes_written)
{
	return 0;
}

int mpc_encoder_encode_tail(mpc_encoder_t instance, mpc_uint8_t *buf, int *bytes_written)
{
	return 0;
}


int mpc_encoder_get_hdr_packet(mpc_encoder_t instance, int type, mpc_uint8_t *buf, int *bytes_written)
{
	mpc_enc_t		*encoder = (mpc_enc_t*)instance;
	mpc_uint8_t		temp[1024];
	mpc_bitstream	bs;

	// output data must be valid
	if (!encoder) return -1;
	if (!buf || !bytes_written) return -1;

	// prepare output bitstream
	mpc_bs_init(&bs, temp);

	switch (type) {
	case PACKET_MAGIC:
		{
			mpc_bs_init(&bs, buf);
			writeMagic(&bs);
			*bytes_written = (bs.buf - temp);
		}
		break;

	case PACKET_TYPE('S','H'):
		{
			writeStreamHeader(encoder, &bs);			
			*bytes_written = writePacket(buf, type, temp, (bs.buf - temp), MPC_TRUE);
		}
		break;

	case PACKET_TYPE('R','G'):
		{
			writeGainInfo(encoder, &bs);			
			*bytes_written = writePacket(buf, type, temp, (bs.buf - temp), MPC_FALSE);
		}
		break;

	case PACKET_TYPE('E','I'):
		{
			writeEncoderInfo(encoder, &bs);			
			*bytes_written = writePacket(buf, type, temp, (bs.buf - temp), MPC_FALSE);
		}
		break;

	// no other packet types
	default:	return -1;
	}

	return 0;
}


