/*
 * Musepack audio compression
 * Copyright (C) 1999-2004 Buschmann/Klemm/Piecha/Wolf
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <math.h>
#include <string.h>

#include <mpc/mpc_types.h>
#include <mpc/datatypes.h>
#include <mpc/mpcmath.h>

#include "../include/libmpcenc.h"
#include "psy_model.h"
#include "bitstream.h"
#include "encoder.h"
#include "syntax_sv8.h"

unsigned long crc32(unsigned char *buf, int len);


void writeMagic(mpc_bitstream *bs)
{
	mpc_bs_putbits(bs, 'M', 8);
	mpc_bs_putbits(bs, 'P', 8);
	mpc_bs_putbits(bs, 'C', 8);
	mpc_bs_putbits(bs, 'K', 8);

	mpc_bs_bytealign(bs, 0);
	mpc_bs_writebits(bs);
}

int writePacket(mpc_uint8_t *outbuf, int code, mpc_uint8_t *payload, int payload_len, mpc_bool_t add_crc)
{
	mpc_bitstream	bs;
	mpc_uint32_t	crc;

	if (!outbuf) return 0;

	mpc_bs_init(&bs, outbuf);
	
	//	All packets are composed as
	//	KEY / SIZE / (optional CRC) / PAYLOAD

	mpc_bs_putbits(&bs, code, 16);

	if (add_crc) {
		mpc_bs_encode_size(&bs, payload_len+4, MPC_TRUE);

		crc = crc32((unsigned char *)payload, payload_len);
		mpc_bs_putbits(&bs, crc, 32);

	} else {
		mpc_bs_encode_size(&bs, payload_len, MPC_TRUE);
	}

	mpc_bs_writebits(&bs);

	// now copy the payload
	if (payload_len > 0) {
		memcpy(bs.buf, payload, payload_len);
		bs.buf += payload_len;
	}

	// return the total length of packet in bytes
	return (int)(bs.buf - outbuf);
}



void writeStreamHeader(mpc_enc_t *e, mpc_bitstream *bs)
{
	mpc_bs_putbits(bs, 8, 8);										// Stream Version

	mpc_bs_encode_size(bs, e->samples_encoded, MPC_FALSE);			// total sample count
	mpc_bs_encode_size(bs, e->samples_to_skip, MPC_FALSE);			// to skip at the beginning
	
	switch (e->config.samplerate) {
	case 44100:		mpc_bs_putbits(bs, 0, 3);	break;
	case 48000:		mpc_bs_putbits(bs, 1, 3);	break;
	case 37800:		mpc_bs_putbits(bs, 2, 3);	break;
	case 32000:		mpc_bs_putbits(bs, 3, 3);	break;
	}

	mpc_bs_putbits(bs, e->psy_model.Max_Band - 1, 5);				// max bandwidth
	mpc_bs_putbits(bs, e->config.channels - 1, 4);					// channels
	mpc_bs_putbits(bs, e->config.ms_channel_mode > 0, 1);			// MS-Coding Flag
	mpc_bs_putbits(bs, e->config.frames_per_block_power >> 1, 3);	// frames per block (log4 unit)

	mpc_bs_bytealign(bs, 0);
	mpc_bs_writebits(bs);
}

void writeGainInfo(mpc_enc_t *e, mpc_bitstream *bs)
{
	mpc_bs_putbits(bs, 1,  8 );										// version
	mpc_bs_putbits(bs, e->config.title_gain,  16 );					// Title gain
	mpc_bs_putbits(bs, e->config.title_peak,  16 );					// Title peak
	mpc_bs_putbits(bs, e->config.album_gain,  16 );					// Album gain
	mpc_bs_putbits(bs, e->config.album_peak,  16 );					// Album peak

	mpc_bs_bytealign(bs, 0);
	mpc_bs_writebits(bs);
}

void writeEncoderInfo(mpc_enc_t *e, mpc_bitstream *bs)
{
	mpc_bs_putbits(bs, (mpc_uint32_t)(e->psy_model.FullQual * 8 + .5), 7);
	mpc_bs_putbits(bs, e->psy_model.PNS > 0, 1);	
	mpc_bs_putbits(bs, MPCENC_MAJOR, 8);	
	mpc_bs_putbits(bs, MPCENC_MINOR, 8);	
	mpc_bs_putbits(bs, MPCENC_BUILD, 8);	

	mpc_bs_bytealign(bs, 0);
	mpc_bs_writebits(bs);
}

void writeSeekTable(mpc_enc_t *e, mpc_bitstream *bs)
{
	// TODO: 
}







